#include <conio.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "instrs.inc"

#define UPPERHALF_FFFF 4294901760
#define LOWERHALF_FFFF 65535
#define MUST_SPECIFY_FILE "You must specify a source file to simulate.\n"
#define USAGE_MESSAGE "Correct Syntax is:\nsim <input file> [output file] [switches]\n"
#define MAX_LINE_LENGTH 254
#define MEM_LOCATIONS_32BIT 262144
#define MAX_NUM_OPS 32768
#define MAX_NUM_LABELS 4096

int displayHex(int value);
bool isLabel(char * token);
int getLabelAddr(char * token);
int processLine(char * line);
int checkReg(int number);
int doArgumentProcessing(int argc, char * argv[]);
int dumpRegs();
int dumpMem();
int strequal(char *a, char *b);
int exit_success();
int print_error(int number);
int setVMessage(char * newMessage);
int parseOpcode(char * token);
int parseRegOrConst(char * token);
int printSyntax();
int processOperation(int opcode, int dest, int src1, int src2);
int doBreak();
int doLabel(char * token);
int doStop();
int readCodeIn();
int executeCode();

struct userdeflabel {
   char labelname[8];
   int addr;
};

struct operation {
   short opcode;
   int src1, src2;
   int dest;
   bool break_after;
};

char vmessage[64];
bool in_comment_zone;
bool reg_dump, mem_dump, verbose, fast, ins_mem_fill, vga_silent;
FILE * outputfile;
FILE * inputfile;

int reg_file[32];
int memory[MEM_LOCATIONS_32BIT];
int ip;
int error_flag;
int next_open_label;
operation ops[MAX_NUM_OPS];
userdeflabel labels[MAX_NUM_LABELS];

int main(int argc, char * argv[]) {
   ip = 0;
   next_open_label = 0;
   error_flag = 0;
   in_comment_zone = false;

   if(doArgumentProcessing(argc,argv)==-1) exit(0);

   while(ip<MAX_NUM_OPS) {
      ops[ip].opcode = -1;
      ip++;
   }

   ip = 0;

   if(!readCodeIn()) executeCode();

   if(reg_dump) dumpRegs();
   if(mem_dump) dumpMem();

   exit_success();
}

int executeCode() {
   if(ins_mem_fill) {
      for(int x = 0; x<ip; x++)
         memory[x] = 0xFFFFFFFF;
   }
   ip = 0;
   int x, y, color;
   int d, s1, s2, brpix_counter, brvid_counter, brchar_counter;
   int reg29;
   brchar_counter = rand();
   brpix_counter = rand();
   brvid_counter = rand();
   bool reg29write;
   bool break_flag = false;
   fprintf(outputfile,"***Execution begins at IP = 0\n\n");
   while(ops[ip].opcode!=-1) {
      reg29 = reg_file[29];
      reg29write = false;
      d = ops[ip].dest;
      s1 = ops[ip].src1;
      s2 = ops[ip].src2;
      reg_file[0] = 0;
      switch(ops[ip].opcode) {
         case ADD:
            reg_file[d] = reg_file[s1] + reg_file[s2];
            reg29write = (d==29);
            ip++;
            break;
         case SUB:
            reg_file[d] = reg_file[s1] - reg_file[s2];
            reg29write = (d==29);
            ip++;
            break;
         case AND:
            reg_file[d] = reg_file[s1] & reg_file[s2];
            reg29write = (d==29);
            ip++;
            break;
         case OR:
            reg_file[d] = reg_file[s1] | reg_file[s2];
            reg29write = (d==29);
            ip++;
            break;
         case NOT:
            reg_file[d] = reg_file[s1] ^ (UPPERHALF_FFFF | LOWERHALF_FFFF);
            reg29write = (d==29);
            ip++;
            break;
         case XOR:
            reg_file[d] = reg_file[s1] ^ reg_file[s2];
            reg29write = (d==29);
            ip++;
            break;
         case LW:
            reg_file[d] = memory[(reg_file[s1]+s2)&(MEM_LOCATIONS_32BIT-1)];
            reg29write = (d==29);
            ip++;
            break;
         case SW:
            memory[(reg_file[s1]+s2)&(MEM_LOCATIONS_32BIT-1)] = reg_file[d];
            ip++;
            break;
         case JR:
            ip = reg_file[d];
            break;
         case BREQ:
            if(reg_file[d]==reg_file[s1]) ip = ip + 1 + s2;
            else ip++;
            break;
         case BRLT:
            if(reg_file[d]<reg_file[s1]) ip = ip + 1 + s2;
            else ip++;
            break;
         case BRGT:
            if(reg_file[d]>reg_file[s1]) ip = ip + 1 + s2;
            else ip++;
            break;
         case NOP:
            ip++;
            break;
         case LLI:
            reg_file[d] = reg_file[d] & UPPERHALF_FFFF;
            s1 = s1 & LOWERHALF_FFFF;
            reg_file[d] = reg_file[d] | s1;
            reg29write = (d==29);
            ip++;
            break;
         case LUI:
            reg_file[d] = reg_file[d] & LOWERHALF_FFFF;
            if(s1<0) s1 = 65536 + s1;
            s1 = s1 * 65536;
            reg_file[d] = reg_file[d] | s1;
            reg29write = (d==29);
            ip++;
            break;
         case BR:
            ip = ip + 1 + d;
            break;
         case SL:
            reg_file[d] = reg_file[s1] * pow(2,(reg_file[s2]%32));
            reg29write = (d==29);
            ip++;
            break;
         case SRL:
            reg_file[d] = ((unsigned int) reg_file[s1]) / ((int) pow(2,(reg_file[s2]%32)));
            reg29write = (d==29);
            ip++;
            break;
         case SRA:
            if(reg_file[s1]<0) {
               reg_file[d] = ((unsigned int) reg_file[s1]) / ((int) pow(2,(reg_file[s2]%32)));
               reg_file[d] = reg_file[d] | ((int) (pow(2,32) - pow(2,(32-reg_file[s2]%32))));
            } else {
               reg_file[d] = reg_file[s1] / ((int) pow(2,(reg_file[s2]%32)));
            }
            reg29write = (d==29);
            ip++;
            break;
         case ROR:
            reg_file[0] = (int) pow(2,(reg_file[s2]%32)) - 1;
            reg_file[0] = reg_file[s1] & reg_file[0];
            reg_file[0] = reg_file[0] * (pow(2,(32-reg_file[s2]%32)));
            if(s1!=0&&s2!=0) reg_file[d] = ((unsigned int) reg_file[s1]) / ((int) pow(2,(reg_file[s2]%32)));
            else if(s1==0) reg_file[d] = 0;
            else reg_file[d] = reg_file[s1];
            reg_file[d] = reg_file[0] | reg_file[d];
            reg29write = (d==29);
            ip++;
            break;
         case ROL:
            if(reg_file[s2]%32!=0) {
            reg_file[0] = (int) pow(2,32)- 1 - ((int) pow(2,(reg_file[s2]%32))-1);
            reg_file[0] = reg_file[s1] & reg_file[0];
            reg_file[0] = ((unsigned int) reg_file[0 ]) / ((int) pow(2,(32-(reg_file[s2]%32))));
            if(s2!=0&&s1!=0) reg_file[d] = reg_file[s1] * pow(2,(reg_file[s2]%32));
            else if(s1==0) reg_file[d] = 0;
            else reg_file[d] = reg_file[s1];
            reg_file[d] = reg_file[0] | reg_file[d]; }
            else reg_file[d] = reg_file[s1];
            reg29write = (d==29);
            ip++;
            break;
         case MUL:
            reg_file[0] = reg_file[s1] * pow(2,16);
            if(s1==0||s2==0) {
               reg_file[d] = 0;
               ip++;
               break;
            }
            if(reg_file[0]<0) {
               reg_file[0] = ((unsigned int) reg_file[0]) / ((int) pow(2,16));
               reg_file[0] = reg_file[0] | ((int) (pow(2,32) - pow(2,16)));
            } else {
               reg_file[0] = reg_file[0] / ((int) pow(2,16));
            }
            reg_file[d] = reg_file[s2] * pow(2,16);
            if(reg_file[d]<0) {
               reg_file[d] = ((unsigned int) reg_file[d]) / ((int) pow(2,16));
               reg_file[d] = reg_file[d] | ((int) (pow(2,32) - pow(2,16))); 
            } else {
               reg_file[d] = reg_file[d] / ((int) pow(2,16));
            }
            reg_file[d] = reg_file[d] * reg_file[0];
            reg29write = (d==29);
            ip++;
            break;
         case INC:
            reg_file[d] = reg_file[d] + 1;
            reg29write = (d==29);
            ip++;
            break;
         case DEC:
            reg_file[d] = reg_file[d] - 1;
            reg29write = (d==29);
            ip++;
            break;
         case JAL:
            reg_file[31] = ip + 1;
            if(d<0) {
               fprintf(outputfile,"JAL address invalid: Put 0 before value: ");
               displayHex(d);
               fprintf(outputfile,"\n");
               exit_success();
            }
            ip = (unsigned) d;
            break;
         case PUSH:
            memory[((unsigned) reg_file[30])&(MEM_LOCATIONS_32BIT-1)] = reg_file[d];
            if((reg_file[30]&0x0003FFFF)!=0) reg_file[30] = reg_file[30] - 1;
            ip++;
            break;
         case POP:
            if((((unsigned int) reg_file[30])&0x0003FFFF)!= 0x0003FFFF) reg_file[30] = reg_file[30] + 1;
            reg_file[d] = memory[((unsigned) reg_file[30])&(MEM_LOCATIONS_32BIT-1)];
            reg29write = (d==29);
            ip++;
            break;
         case SVGA:
            if(brpix_counter==0) {
                  if(!vga_silent) {
                  color = reg_file[d] & 0x000000FF;
                  x = reg_file[d] & 0x0003FF00;
                  x = x / 0x00000100;
                  y = reg_file[d] & 0x0FFC0000;
                  y = y / 0x00040000;
                  fprintf(outputfile,"-->Pixel printed: X=%i\t  Y=%i\t  COLOR=%i\n",x,y,color);
                  }
            } else {
                  fprintf(outputfile,"\n-->SVGA instruction when VGA controller not ready, IP =");
                  displayHex(ip);
                  fprintf(outputfile,"\n");
            }
            ip++;
            break;
         case BRCHAR:
            if(brchar_counter==0) {
               ip = ip + 1 + d;
               if(!fast) brchar_counter = rand();
               else      brchar_counter = 3;
               reg_file[29] = getch();
               if(reg_file[29]==0x00000003) {
                  break_flag = true;
                  break;
               }
               if(reg_file[29]==0x0000000D) reg_file[29] = 0x0000000A;
               if(reg_file[29]==0||reg_file[29]==0xE0) {
                  reg_file[29] = getch();
                  if(reg_file[29]==0x0000004B) {
                     reg_file[29] = 0x00000034;
                  } else if(reg_file[29]==0x00000048) {
                     reg_file[29] = 0x00000038;
                  } else if(reg_file[29]==0x00000050) {
                     reg_file[29] = 0x00000032;
                  } else if(reg_file[29]==0x0000004D) {
                     reg_file[29] = 0x00000036;
                  } 
               }
            } else {
               brchar_counter--;
               ip++;
            }
            break;
         case BRVID:
            if(brvid_counter==1) {
               ip = ip + 1 + d;
               brvid_counter--;
            } else if(brvid_counter==0) {
               if(!fast) brvid_counter = rand();
               else      brvid_counter = 3;
               ip++;
            } else {
               brvid_counter--;
               ip++;
            }
            break;
         case BRPIX:
            if(brpix_counter==1) {
               ip = ip + 1 + d;
               brpix_counter--;
            } else if(brpix_counter==0) {
               if(!fast) brpix_counter = rand();
               else      brpix_counter = 3;
               ip++;
            } else {
               brpix_counter--;
               ip++;
            }                              
            break;
         default:
            ip++;
            break;
      }
      reg_file[29] = reg_file[29] & 0x000000FF;
      if(reg29write)  {
         if(brvid_counter==0) {
            putc(reg_file[29],outputfile);
            if(reg_file[29]==0x00000008) {
               putc(' ',outputfile);
               putc(reg_file[29],outputfile);
            }
         } else {
               fprintf(outputfile,"-->Reg 29 written while VGA controller unready. IP = ");
               displayHex(ip);
               fprintf(outputfile,"\n");
         }
         reg_file[29] = reg29;
      }
      if(break_flag) break;
   }
   reg_file[0] = 0;
   if (break_flag) fprintf(outputfile,"\n\n***User hit CTRL+C. Last IP = ");
   else fprintf(outputfile,"\n\n***Stop reached, IP = ");
   displayHex(ip);
   fprintf(outputfile,"\n");
   return 0;
}

int dumpRegs() {
   int i = 0;
   fprintf(outputfile,"\nDumping Register File:\n");
   while(i<32) {
      if(i<10) fprintf(outputfile,"R%i  = ",i);
      else fprintf(outputfile,    "R%i = ",i);
      displayHex(reg_file[i]);
      i++;
      if(i%3==0) fprintf(outputfile,"\n");
      else fprintf(outputfile,"\t");
   }

   return 0;
}

int dumpMem() {
   int address = 0;
   fprintf(outputfile,"\n\nDumping Memory: Note locations with value = 0 are not dumped.\n");
   while(address<MEM_LOCATIONS_32BIT) {
      if(memory[address] != 0) {
         fprintf(outputfile,"Address: ");
         displayHex(address);
         fprintf(outputfile,"\tValue: ");
         displayHex(memory[address]);
         fprintf(outputfile,"\n");
      }
      address++;
   }
   return 0;
}

int readCodeIn() {
   char line_in[MAX_LINE_LENGTH+1];
   int line_number = 1; // start @ line 1
   int process_result;

   char * token;
   int temp;
   bool in_comment_zone = false;
   while( fgets(line_in,MAX_LINE_LENGTH+1,inputfile) ) {
      if(line_in[0]=='/')
         if(line_in[1]=='/') // comment, skip rest of line
            continue;

      token = strtok(&line_in[0]," \t\n,");
      if(token==NULL) continue;

      temp = 0;
      while(line_in[temp]!=NULL) {
         if(line_in[temp]=='/')
            if(line_in[temp+1]=='*')
               in_comment_zone = true;
         if(line_in[temp]=='*')
            if(line_in[temp+1]=='/')
               in_comment_zone = false;
         temp++;
      }

      token = strtok(&line_in[0]," \n\t,");

      if(!in_comment_zone&&line_in[0]==':') {
         doLabel(token);
         
      } else if(!strequal(token,"stop")) ip++;
      else {
      ip; }
   } // loop for label detection
   ip = 0;

   rewind(inputfile);

   while( fgets(line_in,MAX_LINE_LENGTH+1,inputfile) ) {
      process_result = processLine(&line_in[0]);

      if(process_result<0) {
         fprintf(outputfile,"*** Error encountered, line %i:\n",line_number);
         fprintf(outputfile,"-->");
         fprintf(outputfile,&vmessage[0]);
         fprintf(outputfile,"\n\n");
         exit_success();
      } else if(process_result>0) {
         fprintf(outputfile,">>> Warning, line: %i\n",line_number);
         fprintf(outputfile,"-->");
         fprintf(outputfile,&vmessage[0]);
         fprintf(outputfile,"\n\n");
      }
      
      line_number++;   
   }
   line_number--;

   fprintf(outputfile,"\n***Read in %i lines without syntax errors.\n",line_number);

   if( !feof(inputfile) ) {
      fprintf(stderr,"Error reading from input file.\n");
      return -1;
   }

   return 0;

}

// returns 0 for no special condition
// returns 1 for warning condition
// returns -1 for error condition

int processLine(char * line) {
   if(verbose) {
      fprintf(outputfile, &line[0]);
   }
   if(line[0]=='\n') return 0;
   char * token;
   token = strtok(line," \n\t,");
   int token_number = -1;
   int i;
   int opcode; int dest; int op1, op2; bool process_op = true;
   bool start_of_line = true;
   while(token!=NULL) {
      
      token_number++;
      if(token[0]=='\n') {
         opcode = NOP;
         return processOperation(opcode,0,0,0);
      }
//      if(verbose && !in_comment_zone &&token[0]!='/') {
//         fprintf(outputfile,"%s\t",token);
//         if(token[0]==':') fprintf(outputfile,"\n");
//      }
//      if(token[0] == '/'&&verbose) fprintf(outputfile,"\n");

      if(!in_comment_zone) {

         // check for commented line
         if(token[0]=='/') {
            if(token[1]=='*') { // start of multi-line comment
               fprintf(outputfile,"Comments of style /* */ not allowed.\n");
               exit_success();
               in_comment_zone = true;
               token = strtok(NULL," \n\t,");
               continue;
            } else if(token[1]=='/') {
               if(start_of_line) return 0; // if rest of line is commented, done processing line
               else return processOperation(opcode,dest,op1,op2);
            }
         }

         // get opcode, destinations, and operands
         if(token_number==0) {
            process_op = true;
            if(strequal(token,"break")) return doBreak();
            else if(strequal(token,"stop")) return doStop();
            else if(token[0]==':') return 0;
            else {
               opcode = parseOpcode(token);
               if(opcode < 0) {
                  setVMessage("Unknown opcode");
                  return -1;
               }
            }
         } else if(token_number==1) {
            dest = parseRegOrConst(token);
            if(error_flag<0) return -1;
         } else if(token_number==2) {
            op1 =  parseRegOrConst(token);
            if(error_flag<0) return -1;
         } else if(token_number==3) {
            op2 =  parseRegOrConst(token);
            if(error_flag<0) return -1;
         } else {
            setVMessage("Unrecognized operation.");
            return -1;
         }
          
      } else {
         i = 0;
         while(token[i]!=NULL) {
            if(token[i]=='*') {
               if(token[i+1]=='/') {
                  in_comment_zone = false;
                  process_op = false;
               }
            }
            i++;
         }
      }
      start_of_line = false;
      token = strtok(NULL," \n\t,");
   }
   if(!in_comment_zone && process_op) return processOperation(opcode,dest,op1,op2);
   else return 0;
}

int processOperation(int opcode, int dest, int src1, int src2) {
//   if(verbose) fprintf(outputfile,"\n");

   ops[ip].opcode = opcode;
   ops[ip].break_after = false;

   // check to make sure first operand (dest) is a reg
   if(opcode!=BR&&opcode!=NOP&&opcode!=JAL&&checkReg(dest)!=0&&
      opcode!=BRCHAR&&opcode!=BRVID&&opcode!=BRPIX) return -1;

   // check to make sure second operand (src1) is a reg
   if(opcode<=SW||(opcode>=BREQ&&opcode<=BRGT)||(opcode>=SL&&opcode<=MUL)) {
      if( checkReg( src1 ) !=0) return -1;
   }

   // check to make sure third operand (src2) is a reg
   if(opcode<=OR||opcode==XOR||(opcode>=SL&&opcode<=MUL)&&opcode!=SW&&opcode!=LW) {
      if( checkReg( src2 ) !=0) return -1;
   }

//   if(opcode>BRCHAR||opcode>=BR&&opcode<=BRGT) {
      // convert actual address to offset
  //    dest = dest - ip - 1;
//   }

//   if(opcode>=BRLT&&opcode<=BRGT) src2 = src2 - ip - 1;

   if(opcode==BR||opcode>=BRCHAR) {
      // CHECK APPROPRIATE IMMEDIATE VALUE HERE ON DEST
      if(dest>8191||dest<-8192) {
         setVMessage("Offset out of bounds");
         return -1;
      }
   }

   if(opcode>=BREQ&&opcode<=BRGT||opcode==LW||opcode==SW) {
      if(src2>8191||dest<-8192) {
         setVMessage("Offset out of bounds");
         return -1;
      }
   }

   if(opcode==LLI||opcode==LUI) {
      // CHECK APPROPRIATE IMMEDIATE VALUE HERE ON SRC1
      if(src1<-32768||src1>32767) {
         printf("\nImmediate value is: %i",src1);
         setVMessage("Immediate out of bounds");
         return -1;
      }
   }

   if(opcode==JAL) {
      // CHECK APPROPRIATE IMMEDIATE VALUE HERE ON DEST
      if(dest>131071||dest<-131072) {
         setVMessage("Invalid JAL Address");
         return -1;
      }
   }

   switch(opcode) {
      case ADD:
      case SUB:
      case AND:
      case OR :
      case NOT:
      case XOR:
      case LW :
      case SW :
      case SL:
      case SRL:
      case SRA:
      case ROR:
      case ROL:
      case MUL: ops[ip].dest = dest;
               ops[ip].src1 = src1;
               ops[ip].src2 = src2;
               break;
      case BREQ:
      case BRLT:
      case BRGT:
               ops[ip].dest = dest;
               ops[ip].src1 = src1;
               ops[ip].src2 = src2 - ip - 1;
               break;
      case BR:
      case BRPIX:
      case BRCHAR:
      case BRVID:
               ops[ip].dest = dest - ip - 1;
               ops[ip].src1 = 0;
               ops[ip].src2 = 0;
               break;
      case JR:
      case INC:
      case DEC:
      case PUSH:
      case POP:
      case SVGA: ops[ip].dest = dest;
               ops[ip].src1 = 0;
               ops[ip].src2 = 0;
               break;
      case NOP: ops[ip].dest = 0;
               ops[ip].src1 = 0;
               ops[ip].src2 = 0;
               break;
      default:
               ops[ip].dest = dest;
               ops[ip].src1 = src1;
               ops[ip].src2 = 0;
               break;
   }
   ip++;
   return 0;
}

int checkReg(int number) {
   if(number>31||number<0) {
      setVMessage("Unrecognized register");
      return -1;
   }
   return 0;
}

int doLabel(char * token) {
   token = &token[1];
   int i;
   for(i = 0; i<7 && token[i]!=NULL; i++)
      labels[next_open_label].labelname[i]=token[i];
   labels[next_open_label].labelname[i]=NULL;
   labels[next_open_label].addr = ip;
   next_open_label++;
   return 0;
}

int getLabelAddr(char * token) {
   int i = 0;
   for(i;i<next_open_label;i++)
      if(strequal(token,&labels[i].labelname[0])) return labels[i].addr;
   return -1;
}

bool isLabel(char * token) {
   int i = 0;
   int j;
   for(i;i<next_open_label;i++) {
      for(j=0;j<7&&token[j]!=NULL;j++) {
         if(token[j]!=labels[i].labelname[j]) break;
      }
      if(token[j]==labels[i].labelname[j]) return true;
   }
   return false;
}

int doBreak() {
   if(ip>0)
   ops[ip-1].break_after = true;
   fprintf(outputfile,"\n");
   return 0;
}

int doStop() {
   ops[ip].break_after = true;
   ops[ip].opcode = -1;
   fprintf(outputfile,"\n");
   return 0;                  
}

int parseOpcode(char * token) {
   if(strequal(token,"add")) {
      return ADD;
   } else if(strequal(token,"sub")) {
      return SUB;
   } else if(strequal(token,"and")) {
      return AND;
   } else if(strequal(token,"or")) {
      return OR;
   } else if(strequal(token,"not")) {
      return NOT;
   } else if(strequal(token,"xor")) {
      return XOR;
   } else if(strequal(token,"lw")) {
      return LW;
   } else if(strequal(token,"sw")) {
      return SW;
   } else if(strequal(token,"jr")) {
      return JR;
   } else if(strequal(token,"breq")) {
      return BREQ;
   } else if(strequal(token,"brlt")) {
      return BRLT;
   } else if(strequal(token,"brgt")) {
      return BRGT;
   } else if(strequal(token,"nop")) {
      return NOP;
   } else if(strequal(token,"lui")) {
      return LUI;
   } else if(strequal(token,"lli")) {
      return LLI;
   } else if(strequal(token,"br")) {
      return BR;
   }else if(strequal(token,"sl")) {
      return SL;
   } else if(strequal(token,"srl")) {
      return SRL;
   } else if(strequal(token,"sra")) {
      return SRA;
   } else if(strequal(token,"ror")) {
      return ROR;
   } else if(strequal(token,"rol")) {
      return ROL;
   } else if(strequal(token,"mul")) {
      return MUL;
   } else if(strequal(token,"inc")) {
      return INC;
   } else if(strequal(token,"dec")) {
      return DEC;
   } else if(strequal(token,"jal")) {
      return JAL;
   } else if(strequal(token,"push")) {
      return PUSH;
   } else if(strequal(token,"pop")) {
      return POP;
   } else if(strequal(token,"svga")) {
      return SVGA;
   } else if(strequal(token,"brchar")) {
      return BRCHAR;
   } else if(strequal(token,"brvid")) {
      return BRVID;
   } else if(strequal(token,"brpix")) {
      return BRPIX;
   }
                   
   return -1;
}

int parseRegOrConst(char * token) {
   if(token==NULL) return 0; // invalid value, take a guess
   int accumulator = 0; char c; int i = 0; int radix = 10;
   bool neg = false;
   if(token[0]=='r'||token[0]=='R'||token[0]=='$') { // its a reg
      i++;
      while(token[i]!=NULL) {
         c = token[i];
         if(c>='0'&&c<='9') c = c - '0';
         else {
            setVMessage("Unknown register or constant");
            error_flag = -1;
            return accumulator;
         }
         accumulator = accumulator * radix + c;
         i++;
      }
   } else { // constant value
     while(token[i]!=NULL) {
      if(token[i]>='a'&&token[i]<='f' ||
         token[i]>='A'&&token[i]<='F')
         radix = 16;
      i++;
     }
     if(i==0) return -34;
     if(token[i-1]=='h'||token[i-1]=='H') {
         radix = 16;
         token[i-1]=NULL;
         i--;
     }
     int length = i;
     i = 0;
     if(token[0]=='-') { i++; }
     if(isLabel(token)) return getLabelAddr(token);
     while(token[i]!=NULL) {
         c = token[i];
         if(c<='f'&&c>='a') {c = c - 'a' + 10; radix = 16;}
         else if(c<='F'&&c>='A') {c = c - 'A' + 10; radix = 16;}
         else if(c>='0'&&c<='9') c = c - '0';
         else {
            setVMessage("Unknown register or constant");
            error_flag = -1;
            return accumulator;
         }
         accumulator = accumulator * radix + c;
         i++;
     }
     if((token[0]>='8'&&token[0]<='9' ||
         token[0]<='f'&&token[0]>='a' ||
         token[0]<='F'&&token[0]>='A' ) && radix==16 &&
         length == 4) neg = true;
     if(neg) accumulator = accumulator - 65536;
     else if(token[0]=='-') accumulator = 0 - accumulator;
   }
   return accumulator;
}

int setVMessage(char * newMessage) {
   if(newMessage==NULL) return -1;
   int i = 0;
   while(i<63 && newMessage[i]!=NULL) {
      vmessage[i] = newMessage[i];
      i++;
   }
   vmessage[i] = NULL;
   return 0;
}

int doArgumentProcessing(int argc, char * argv[]) {
   if(argc == 1) {
      fprintf(stderr,MUST_SPECIFY_FILE);
      printSyntax();
      return -1;
   }

   inputfile = fopen(argv[1],"r");
   if(inputfile ==NULL) {
      fprintf(stderr,"Cannot open source file: %s\n",argv[1]);
      return -1;
   }

   outputfile = stdout;
   bool output_to_file = false;
   reg_dump = mem_dump = output_to_file;
   verbose = true;
   ins_mem_fill = false;
   vga_silent = false;
   fast = false;

   for(int i=2;i<argc;i++) {
      if(strequal(argv[i],"+r")) reg_dump = true;
      else if(strequal(argv[i],"+m")) mem_dump = true;
      else if(strequal(argv[i],"-v")) verbose = false;
      else if(strequal(argv[i],"+h")) return printSyntax();
      else if(strequal(argv[i],"+i")) ins_mem_fill = true;
      else if(strequal(argv[i],"+vgasilent")) vga_silent = true;
      else if(strequal(argv[i],"+fast")) fast = true;
      else if(output_to_file==false) {
         output_to_file = true;
         outputfile = fopen(argv[i],"w");
         if(outputfile==NULL) {
            fprintf(stderr,"Cannot open file for writing: %s\n",argv[i]);
            return -1;
         }

         fprintf(stdout,"Using output file: %s\n",argv[i]);

      } else {
         fprintf(stderr,"The following argument is not recognized: %s\n",argv[i]);
         return -1;
      }
   }

   return 0;
}

int strequal(char * a, char * b) {
   if(stricmp(a,b)==0) return 1;
   else return 0;
}

int exit_success() {
   fclose(outputfile);
   fclose(inputfile);
   exit(0);
   return -1;
}

int displayHex(int value) {
   unsigned int x = value;
   char out[16];
   int temp;
   for(int i=0;i<16;i+=2) {
      temp = (x % 16);
      if(temp<=9) out[i] = temp + '0';
      else out[i] = temp + 'A' - 10;
      out[i+1]=NULL;
      x = x / 16;
   }
   for(int i=14;i>=0;i-=2) 
      fprintf(outputfile,&out[i]);
   
   return 0;
}

int printSyntax() {
   fprintf(stderr,USAGE_MESSAGE);
   fprintf(stderr,"Switches:\n+h  Display this message\n+m  Dump non-zero memory locations after execution\n+r  Dump register file after execution\n-v  Verbose mode off (don't echo code)\n");
   fprintf(stderr,"+i fill instruction memory with 0xFFFFFFFF\n+vgasilent Don't print pixel manipulation information\n");
   fprintf(stderr,"+fast do I/O without spinning. Useful for many pixel/char manipulations\n");
   return -1;
}             
